#import "Fiend.h"

RCS_ID ("$Id$");

#import "FILevelRepresentation.h"

#import "FIAppInfo.h"
#import "FIApplication.h"
#import "FILevel.h"
#import "FITile.h"
#import "FITileWindow.h"
#import "FIWindowController.h"

@implementation FILevelRepresentation

- initWithLevel:(FILevel *)aLevel
{
    NSEnumerator *applicationEnumerator;
    FIApplication *application;
    FITile *tile;
    NSNotificationCenter *defaultCenter;
    
    if ([super init] == nil)
        return nil;

    nonretainedWindowController = nil;
    level = [aLevel retain];
    tiles = [[NSMutableArray allocWithZone:[self zone]] initWithCapacity:8];
    isVisible = NO;

    applicationEnumerator = [[level applications] objectEnumerator];
    while (application = [applicationEnumerator nextObject])
    {
        tile = [[FITile allocWithZone:[self zone]] initWithApplication:application];
        //[tiles addObject:tile];
        [self addTile:tile checkActive:NO];
        [tile release];
    }

    defaultCenter = [NSNotificationCenter defaultCenter];
    [defaultCenter addObserver:self
                   selector:@selector (levelDidAddApplication:)
                   name:FILevelDidAddApplicationNotification
                   object:level];
    [defaultCenter addObserver:self
                   selector:@selector (levelDidRemoveApplication:)
                   name:FILevelDidRemoveApplicationNotification
                   object:level];

    return self;
}

//----------------------------------------------------------------------

- (void) dealloc
{
    [[NSNotificationCenter defaultCenter] removeObserver:self];
    
    SNRelease (tiles);
    SNRelease (level);

    [super dealloc];
}

//======================================================================
// Controlling general attributes
//======================================================================

- (FILevel *) level
{
    return level;
}

//----------------------------------------------------------------------

- (BOOL) isVisible
{
    return isVisible;
}

//----------------------------------------------------------------------

- (void) setIsVisible:(BOOL)flag
{
    if (flag == isVisible)
        return;

    isVisible = flag;
    if (isVisible == YES)
        [self unhideLevel];
    else
        [self hideLevel];
}

//----------------------------------------------------------------------

- (FIWindowController *) windowController
{
    return nonretainedWindowController;
}

//----------------------------------------------------------------------

- (void) setWindowController:(FIWindowController *)newWindowController
{
    nonretainedWindowController = newWindowController;
    [tiles makeObjectsPerformSelector:@selector (setDelegate:) withObject:newWindowController];
}

//======================================================================
// Debugging
//======================================================================

- (NSMutableDictionary *) debugDictionary
{
    NSMutableDictionary *debugDictionary;

    debugDictionary = [super debugDictionary];
    if (level != nil)
        [debugDictionary setObject:level forKey:@"level"];
    if (tiles != nil)
        [debugDictionary setObject:tiles forKey:@"tiles"];

    return debugDictionary;
}

//======================================================================
// Handling Workspace notifications
//======================================================================

- (void) applicationWillLaunch:(NSNotification *)aNotification
{
    [tiles makeObjectsPerformSelector:@selector (applicationWillLaunch:) withObject:aNotification];
}

//----------------------------------------------------------------------

- (void) applicationDidLaunch:(NSNotification *)aNotification
{
    [tiles makeObjectsPerformSelector:@selector (applicationDidLaunch:) withObject:aNotification];
}

//----------------------------------------------------------------------

- (void) applicationDidTerminate:(NSNotification *)aNotification
{
    [tiles makeObjectsPerformSelector:@selector (applicationDidTerminate:) withObject:aNotification];
}

//======================================================================
// Controlling tiles
//======================================================================

- (NSArray *) tiles
{
    return tiles;
}

//----------------------------------------------------------------------

- (FITile *) tileForApplication:(FIApplication *)application
{
    NSEnumerator *tileEnumerator;
    FITile *tile = nil;

    tileEnumerator = [tiles objectEnumerator];
    while (tile = [tileEnumerator nextObject])
    {
        //NSLog (@"app: %@", [tile application]);
        //NSLog (@"target: %p, this: %p", application, [tile application]);
        if ([tile application] == application)
            break;
    }

    return tile;
}

//----------------------------------------------------------------------

- (FITile *) tileAtDockLocation:(DockLocation)location
{
    NSEnumerator *tileEnumerator;
    FITile *tile, *targetTile;
    DockLocation currentLocation;

    targetTile = nil;
    tileEnumerator = [tiles objectEnumerator];
    while (tile = [tileEnumerator nextObject])
    {
        currentLocation = [[tile application] location];
        if (currentLocation.x == location.x && currentLocation.y == location.y)
        {
            targetTile = tile;
            break;
        }
    }

    return targetTile;
}

//----------------------------------------------------------------------

- (void) addTile:(FITile *)tile checkActive:(BOOL)checkActive
{
    OBPRECONDITION (tile != nil);

    [tiles addObject:tile];
    [tile setDelegate:[self windowController]];
    if (checkActive == YES)
        [tile checkForActiveApplications:[FIAppInfo activeAppInfo]]; // XXX: This may be a little heavy weight..
    [[tile tileWindow] setLevel:[[self windowController] tileWindowLevel]];
    [tile setIsVisible:isVisible];
}

//----------------------------------------------------------------------

- (void) _removeTile:(FITile *)tile
{
    OBPRECONDITION (tile != nil);
    OBPRECONDITION ([tiles containsObject:tile] == YES);

    [tiles removeObject:tile];
}

//----------------------------------------------------------------------

- (void) removeTile:(FITile *)tile
{
    [tile setIsVisible:NO];
    [tile stopTimers];
    [self _removeTile:tile];
}

//======================================================================
// Handling model notifications
//======================================================================

- (void) levelDidAddApplication:(NSNotification *)aNotification
{
    NSDictionary *userInfo;
    FIApplication *application;
    FITile *tile;
    
    userInfo = [aNotification userInfo];
    application = [userInfo objectForKey:@"application"];
    tile = [[FITile allocWithZone:[self zone]] initWithApplication:application];
#if 0
    [self addTile:tile checkActive:YES];
#else
    // This approximately doubles the speed.  The rest of the time is going to getting
    // the file icon from the workspace.
    [self addTile:tile checkActive:NO];
    [NSObject cancelPreviousPerformRequestsWithTarget:[self windowController]
              selector:@selector (checkForActiveApplications:) object:self];
    [[self windowController] performSelector:@selector (checkForActiveApplications:) withObject:self afterDelay:0];
#endif
    // Take on the visibility status for the level representation in case we undo a cut after
    // switch levels.
    [tile setIsVisible:isVisible];
    [tile release];
}

//----------------------------------------------------------------------

- (void) levelDidRemoveApplication:(NSNotification *)aNotification
{
    NSDictionary *userInfo;
    FIApplication *application;
    FITile *tile;
    
    userInfo = [aNotification userInfo];
    application = [userInfo objectForKey:@"application"];
    tile = [self tileForApplication:application];
    if ([tile isSelected] == YES)
        [[self windowController] deselectTile:tile];

    [self removeTile:tile];
}

//======================================================================
// Controlling tile windows
//======================================================================

- (void) setTileWindowLevel:(int)windowLevel
{
    NSEnumerator *tileEnumerator;
    FITile *tile;

    tileEnumerator = [tiles objectEnumerator];
    while (tile = [tileEnumerator nextObject])
    {
        // Huh?  Why do we need the typecast?
        [(NSWindow *)[tile tileWindow] setLevel:windowLevel];
    }
}

//----------------------------------------------------------------------

- (void) hideLevel
{
    [tiles makeObjectsPerformSelector:@selector (hide)];
}

//----------------------------------------------------------------------

- (void) unhideLevel
{
    // Do this lazily for quicker startup, but all at once before we unhide.
    [tiles makeObjectsPerformSelector:@selector (updateFileIconIfNeeded)];
    [tiles makeObjectsPerformSelector:@selector (unhide)];
}

//----------------------------------------------------------------------

- (void) repositionRelativeToPoint:(NSPoint)origin
{
    NSEnumerator *tileEnumerator;
    FITile *tile;
    //DockLocation location;

    tileEnumerator = [tiles objectEnumerator];
    while (tile = [tileEnumerator nextObject])
    {
#if 0
        location = [tile location];
        if (location.x != 0 || location.y != 0)
#endif
            [tile repositionRelativeToPoint:origin];
    }
}

//======================================================================
// Other
//======================================================================

- (void) checkForActiveApplications:(NSDictionary *)activeApplications
{
    [tiles makeObjectsPerformSelector:@selector (checkForActiveApplications:) withObject:activeApplications];
}

//----------------------------------------------------------------------

- (void) performAutoLaunch
{
    [tiles makeObjectsPerformSelector:@selector (performAutoLaunch)];
}

//----------------------------------------------------------------------

- (void) updateFileIcons
{
    [tiles makeObjectsPerformSelector:@selector (updateFileIcon)];
}

//----------------------------------------------------------------------

- (void) updateFileIconsIfNeeded
{
    [tiles makeObjectsPerformSelector:@selector (updateFileIconIfNeeded)];
}

@end
