#import "Fiend.h"

RCS_ID ("$Id$");

#import "FITile.h"

#import "Fiendish.h"
#import "FIApplication.h"
#import "FIAppInfo.h"
#import "FILauncher.h"
#import "FILevel.h"
#import "FITileView.h"
#import "FITileWindow.h"
#import "FITileWindowClone.h"
#import "FISoundManager.h"
#import "FIDock.h"

DEFINE_NSSTRING (FITileIconDidChangeNotification);

NSString *NSStringFromApplicationState (ApplicationState state)
{
    NSString *str;
    
    switch (state)
    {
      case ApplicationInactiveState:
          str = NSSTRINGIFY (ApplicationInactiveState);
          break;
          
      case ApplicationLaunching1State:
          str = NSSTRINGIFY (ApplicationLaunching1State);
          break;
          
      case ApplicationLaunching2State:
          str = NSSTRINGIFY (ApplicationLaunching2State);
          break;
          
      case ApplicationActiveState:
          str = NSSTRINGIFY (ApplicationActiveState);
          break;
          
      default:
          str = @"<unknown application state>";
    }

    return str;
}

//======================================================================

@implementation FITile

- initWithApplication:(FIApplication *)anApplication
{
    NSRect contentFrame;
    NSNotificationCenter *defaultCenter;
    
    if ([super init] == nil)
        return nil;

    application = [anApplication retain];
    tileWindow = [[FITileWindow allocWithZone:[self zone]] init];
    [tileWindow setDelegate:self];

    contentFrame = [[tileWindow contentView] frame];
    tileView = [[FITileView allocWithZone:[self zone]] initWithFrame:contentFrame];
    [tileView setBoundsSize:NSMakeSize (64, 64)];
    [tileWindow setContentView:tileView];

    [tileView setDatasource:self];
    [tileView setDelegate:self];

    hasLoadedFileIcon = NO;
    fileIcon = nil;
    isVisible = NO;
    isSelected = NO;
    tileLaunchedApplication = NO;

    state = ApplicationInactiveState;
    appInfoTimer = nil;
    appInfo = nil;
    hasCheckedIfActive = NO;

    animationTimer = nil;
    animationImage = nil;
    autoLaunchTimer = nil;

    nonretainedDelegate = nil;

    [tileView setNeedsDisplay:YES];

    defaultCenter = [NSNotificationCenter defaultCenter];
    [defaultCenter addObserver:self
                   selector:@selector (applicationDidChangeDockLocation:)
                   name:FIApplicationLocationDidChangeNotification
                   object:application];

    [defaultCenter addObserver:self
                   selector:@selector (applicationDidChange:)
                   name:FIApplicationDidChangeNotification
                   object:application];

    return self;
}

//----------------------------------------------------------------------

- (void) dealloc
{
    [[NSNotificationCenter defaultCenter] removeObserver:self];

    [self unhideSourceTile];

    [self stopAnimationTimer];
    [self stopAppInfoTimer];
    [self stopAutoLaunchTimer];
    
    SNRelease (appInfo);
    SNRelease (fileIcon);
    SNRelease (tileView);
    SNRelease (tileWindow);
    SNRelease (application);

    [super dealloc];
}

//----------------------------------------------------------------------
//" The timers reatin the tile...
//----------------------------------------------------------------------

- (void) stopTimers
{
    [self stopAppInfoTimer];
    [self stopAnimationTimer];
    [self stopAutoLaunchTimer];
}

//======================================================================
// Controlling the delegate
//======================================================================

- delegate
{
    return nonretainedDelegate;
}

//----------------------------------------------------------------------

- (void) setDelegate:newDelegate
{
    nonretainedDelegate = newDelegate;
    [nonretainedDelegate repositionTile:self]; // XXX: Should we do this here?
}

//======================================================================
// Controlling general attributes
//======================================================================

- (FIApplication *) application
{
    return application;
}

//----------------------------------------------------------------------

- (FIDock *) dock
{
  return [[[self application] level] dock];
}

//----------------------------------------------------------------------

- (BOOL) isSelected
{
    return isSelected;
}

//----------------------------------------------------------------------

- (void) setIsSelected:(BOOL)flag
{
    if (flag == isSelected)
        return;

    isSelected = flag;
    [tileView setNeedsDisplay:YES];
}

//----------------------------------------------------------------------

- (BOOL) hasCheckedIfActive
{
    return hasCheckedIfActive;
}

//======================================================================
// Handling state changes
//======================================================================

- (void) enterInactiveState
{
    switch (state)
    {
      case ApplicationLaunching1State:
          OBPRECONDITION (appInfo == nil);
          OBPRECONDITION (appInfoTimer == nil);
          break;
          
      case ApplicationLaunching2State:
          OBPRECONDITION (appInfo == nil);
          [self stopAppInfoTimer];
          break;
          
      case ApplicationActiveState:
          OBPRECONDITION (appInfoTimer == nil);
          SNRelease (appInfo);
          [self stopAnimationTimer];
          break;
          
      case ApplicationInactiveState:
      default:
          OBPRECONDITION (appInfo == nil);
          OBPRECONDITION (appInfoTimer == nil);
          break;
    }

    tileLaunchedApplication = NO;
    state = ApplicationInactiveState;
    [tileView setNeedsDisplay:YES];
}

//----------------------------------------------------------------------

- (void) enterLaunching1State
{
    NSAssert2 (state == ApplicationInactiveState || state == ApplicationLaunching1State,
               @"%@:[%@] Expected to be in ApplicationInactiveState or ApplicationLaunching1State.",
               [application shortName], NSStringFromApplicationState (state));
    OBPRECONDITION (appInfo == nil);
    OBPRECONDITION (appInfoTimer == nil);

    state = ApplicationLaunching1State;
    [tileView setNeedsDisplay:YES];
}

//----------------------------------------------------------------------

- (void) enterLaunching2State
{
    // We don't appear to be getting the -applicationWillLaunch: notifications.
    NSAssert2 (state == ApplicationLaunching1State || state == ApplicationInactiveState,
               @"%@:[%@] Expected to be in either ApplicationInactiveState or ApplicationLaunching1State.",
               [application shortName], NSStringFromApplicationState (state));
    OBPRECONDITION (appInfo == nil);
    OBPRECONDITION (appInfoTimer == nil);

    state = ApplicationLaunching2State;
    [self startAppInfoTimer];
    [tileView setNeedsDisplay:YES];
}

//----------------------------------------------------------------------

- (void) enterActiveState:(FIAppInfo *)anAppInfo
{
    NSAssert2 (state == ApplicationInactiveState || state == ApplicationLaunching2State,
               @"%@:[%@] Expected to be in either ApplicationInactiveState or ApplicationLaunching2State.",
               [application shortName], NSStringFromApplicationState (state));
    OBPRECONDITION (appInfo == nil);

    if (state == ApplicationInactiveState)
    {
        OBPRECONDITION (appInfoTimer == nil);
    }

    [self stopAppInfoTimer];
    appInfo = [anAppInfo retain];

    [self hideSourceTile];

    state = ApplicationActiveState;
    [tileView setNeedsDisplay:YES];
    if ([application shouldAnimate] == YES)
        [self startAnimationTimer];

    if (tileLaunchedApplication == YES)
    {
        [appInfo activateApplicationAndHideOthers:NO];
        tileLaunchedApplication = NO;
    }
#if 0
    // Simple hackery.
    if ([path isEqual:[[FILauncher sharedLauncher] lastLaunchPath]] == YES)
    {
        [appInfo activateApplicationAndHideOthers:NO];
        [[FILauncher sharedLauncher] setLastLaunchPath:nil];
    }
#endif
}

//======================================================================
// Managing app info timer
//======================================================================

- (void) startAppInfoTimer
{
    OBPRECONDITION (appInfoTimer == nil);

    appInfoTimer = [[NSTimer scheduledTimerWithTimeInterval:0.2
                             target:self
                             selector:@selector (fireAppInfoTimer:)
                             userInfo:nil
                             repeats:YES] retain];
}

//----------------------------------------------------------------------

- (void) stopAppInfoTimer
{
    if (appInfoTimer != nil)
    {
        [appInfoTimer invalidate];
        SNRelease (appInfoTimer);
    }
}

//----------------------------------------------------------------------

- (void) fireAppInfoTimer:(NSTimer *)aTimer
{
    NSDictionary *activeApplications;
    FIAppInfo *anAppInfo;

    // XXX: This is possibly a little heavy weight.  Can we have a centralized timer
    // so that we can share the active application info?

    activeApplications = [FIAppInfo activeAppInfo];
    anAppInfo = [activeApplications objectForKey:[application shortName]];
    if (anAppInfo != nil)
    {
        //NSLog (@"Got it! %@", anAppInfo);
        [self enterActiveState:anAppInfo];
    }
    else
    {
        //NSLog (@"Couldn't get app info for '%@'.", [application shortName]);
    }
}

//======================================================================
// Managing animation timer
//======================================================================

- (void) startAnimationTimer
{
    double animationInterval;
    
    OBPRECONDITION (animationTimer == nil);

    animationInterval = [[OFUserDefaults sharedUserDefaults] floatForKey:DK_AnimationInterval];
    animationTimer = [[NSTimer scheduledTimerWithTimeInterval:animationInterval
                               target:self
                               selector:@selector (fireAnimationTimer:)
                               userInfo:nil
                               repeats:YES] retain];
}

//----------------------------------------------------------------------

- (void) stopAnimationTimer
{
    if (animationTimer != nil)
    {
        [animationTimer invalidate];
        SNRelease (animationTimer);
    }

    SNRelease (animationImage);
    [tileView setNeedsDisplay:YES];
}

//----------------------------------------------------------------------

- (void) fireAnimationTimer:(NSTimer *)aTimer
{
    int status;
    
    if (appInfo != nil)
    {
        if ([appInfo tokenWindowNumber] < 0)
        {
            // i.e. Fiend itself has window number of "-1"...
            [self stopAnimationTimer];
        }
        else
        {
            if (animationImage == nil)
            {
                animationImage = [[NSImage allocWithZone:[self zone]] initWithSize:NSMakeSize (60, 60)];
                //[tileView setAnimationImage:animationImage];
                [tileView setNeedsDisplay:YES];
            }

            [animationImage lockFocus];
            // Yes, this is (2,1), not (2, 2)...
            status = safeCopyWindow (2, 1, 60, 60, [appInfo tokenWindowNumber]);
            [animationImage unlockFocus];

            if (status != 0)
            {
                NSLog (@"safeCopyWindow() failed.");
            }
            else
            {
                [tileView setNeedsDisplay:YES];
            }
        }
    }
}

//======================================================================
// Managing delayed auto launch timer
//======================================================================

- (void) startAutoLaunchTimer
{
    double delay;
    
    OBPRECONDITION (autoLaunchTimer == nil);
    delay = secondsInTimeInterval ([application autoLaunchDelay]);
    OBASSERT (delay > 0);

    autoLaunchTimer = [[NSTimer scheduledTimerWithTimeInterval:delay
                                target:self
                                selector:@selector (fireAutoLaunchTimer:)
                                userInfo:nil
                                repeats:NO] retain];
}

//----------------------------------------------------------------------

- (void) stopAutoLaunchTimer
{
    if (autoLaunchTimer != nil)
    {
        [autoLaunchTimer invalidate];
        SNRelease (autoLaunchTimer);
    }
}

//----------------------------------------------------------------------

- (void) fireAutoLaunchTimer:(NSTimer *)aTimer
{
    [self stopAutoLaunchTimer];
    [self enterLaunching1State]; // XXX: New.
    [[FILauncher sharedLauncher] launchApplication:[application path] autolaunch:YES];
}

//======================================================================
// Controlling the source tile
//======================================================================

//----------------------------------------------------------------------
//" If the system automatically creates application icons (RDR2 doesn't),
//" we want to move it out of the way.
//----------------------------------------------------------------------

- (void) hideSourceTile
{
    if ([[OFUserDefaults sharedUserDefaults] boolForKey:DK_HideTiles] == YES)
        [appInfo hideTokenWindow];
}

//----------------------------------------------------------------------
//" If we hide the application icons, then when we quit we need to
//" put them back on the screen.  If we crash they will remain hidden.
//----------------------------------------------------------------------

- (void) unhideSourceTile
{
    if ([[OFUserDefaults sharedUserDefaults] boolForKey:DK_HideTiles] == YES)
        [appInfo unhideTokenWindow];
}


//======================================================================
// FITileView datasource
//======================================================================

- (NSImage *) animationImage
{
    return animationImage;
}

//----------------------------------------------------------------------

- (BOOL) shouldAutoLaunch
{
    return [application shouldAutoLaunch];
}

//----------------------------------------------------------------------

- (BOOL) shouldAnimate
{
    return [application shouldAnimate];
}

//----------------------------------------------------------------------

- (BOOL) isPersistent
{
    return [application isPersistent];
}

//----------------------------------------------------------------------

- (BOOL) isRunning
{
    return (state == ApplicationLaunching1State) || (state == ApplicationLaunching2State) || (state == ApplicationActiveState);
}

//----------------------------------------------------------------------

- (BOOL) isGhosted
{
    return (state == ApplicationLaunching1State) || (state == ApplicationLaunching2State);
}

//======================================================================
// FITileView delegate
//======================================================================

#define HYSTERESIS_VALUE (8)

- (void) tileView:(FITileView *)aTileView singleClicked:(NSEvent *)theEvent
{
    NSPoint originalWindowOrigin, originalScreenPoint;
    NSPoint startPoint;
    NSRect hysteresisRect;

    originalWindowOrigin = [tileWindow frame].origin;
    originalScreenPoint = [tileWindow convertBaseToScreen:[theEvent locationInWindow]];

    //SNLogPoint (originalWindowOrigin);
    //SNLogPoint (originalScreenPoint);

    // First, gobble up any mouse dragged events within a hysteresis rect.

    startPoint = [tileView convertPoint:[theEvent locationInWindow] fromView:nil];
    hysteresisRect.origin.x = startPoint.x - HYSTERESIS_VALUE;
    hysteresisRect.origin.y = startPoint.y - HYSTERESIS_VALUE;
    hysteresisRect.size.width = 2 * HYSTERESIS_VALUE;
    hysteresisRect.size.height = 2 * HYSTERESIS_VALUE;

    theEvent = [self trackMouseInRect:hysteresisRect];

    if ([theEvent type] == NSLeftMouseDragged)
    {
        DockLocation locationOffset;
        FITileWindowClone *ghostWindow, *cloneWindow;
        float tileWidth;

        BOOL stopTracking;
        NSPoint mouseDownPoint;
        NSRect currentRect;
        NSPoint mouseLocation;
        NSPoint newWindowOrigin;
        DockLocation appLocation, ghostLocation;

        tileWidth = [[OFUserDefaults sharedUserDefaults] floatForKey:DK_TileSize];
        appLocation = [application location];

        ghostWindow = [[FITileWindowClone allocWithZone:[self zone]] initWithWindow:tileWindow isGhosted:YES];
        cloneWindow = [[FITileWindowClone allocWithZone:[self zone]] initWithWindow:tileWindow isGhosted:NO];
        [ghostWindow orderFront:self];
        [cloneWindow orderFront:self];
        [tileWindow orderOut:self];

        mouseDownPoint = [tileWindow convertBaseToScreen:[theEvent locationInWindow]];
        mouseDownPoint.x -= tileWidth / 2;
        mouseDownPoint.y -= tileWidth / 2;
        currentRect.origin = mouseDownPoint;
        currentRect.size = NSMakeSize (tileWidth, tileWidth);
        //----------------------------------------
            
        stopTracking = NO;
        while (stopTracking == NO)
        {
            mouseLocation = [tileWindow convertBaseToScreen:[theEvent locationInWindow]];
            //SNLogPoint (mouseLocation);
            locationOffset.x = floor ((mouseLocation.x - mouseDownPoint.x) / tileWidth);
            locationOffset.y = floor ((mouseLocation.y - mouseDownPoint.y) / tileWidth);
            currentRect.origin.x = mouseDownPoint.x + locationOffset.x * tileWidth;
            currentRect.origin.y = mouseDownPoint.y + locationOffset.y * tileWidth;
            //NSLog (@"offset: (%d,%d)", locationOffset.x, locationOffset.y);

            ghostLocation.x = appLocation.x + locationOffset.x;
            ghostLocation.y = appLocation.y + locationOffset.y;
            if (ghostLocation.x == 0 && ghostLocation.y == 0)
            {
                [ghostWindow setFrameOrigin:originalWindowOrigin];
                [ghostWindow setShowCrossbones:YES];
            }
            else
            {
                [ghostWindow setShowCrossbones:NO];
                [nonretainedDelegate tile:self draggedGhostWindow:ghostWindow toLocation:ghostLocation];
            }

            //----------
            while ([theEvent type] == NSLeftMouseDragged)
            {
                mouseLocation = [tileWindow convertBaseToScreen:[theEvent locationInWindow]];
                //SNLogPoint (mouseLocation);
                if (NSPointInRect (mouseLocation, currentRect) == NO)
                    break;

                newWindowOrigin.x = originalWindowOrigin.x + mouseLocation.x - originalScreenPoint.x;
                newWindowOrigin.y = originalWindowOrigin.y + mouseLocation.y - originalScreenPoint.y;

                [cloneWindow setFrameOrigin:newWindowOrigin];

                theEvent = [tileWindow nextEventMatchingMask:(NSLeftMouseUpMask | NSLeftMouseDraggedMask)];
            }

            if ([theEvent type] == NSLeftMouseUp)
                stopTracking = YES;
            else
                theEvent = [tileWindow nextEventMatchingMask:(NSLeftMouseUpMask | NSLeftMouseDraggedMask)];
            //----------
        }

        //----------------------------------------
        //NSLog (@"location offset: %@", NSStringFromDockLocation (locationOffset));
        //NSLog (@"ghost location: %@", NSStringFromDockLocation (ghostLocation));
        if (ghostLocation.x == 0 && ghostLocation.y == 0)
        {
            // Remove tile from dock.
	    [[FISoundManager sharedSoundManager] didRemoveTilesForDock:[self dock]];
            [[application level] removeApplication:application];
            [[application undoManager] setActionName:@"Remove Application"];
        }
        else
        {
            // Set tile location to ghostLocation.
            [application setLocation:ghostLocation];
            [[application  undoManager] setActionName:@"Move Tile"];
        }

        //[tileWindow setFrameOrigin:[ghostWindow frame].origin]; // Should just listen for notification
        //[self didDragToOffset:locationOffset];
        [tileWindow orderFront:self];
        [cloneWindow orderOut:self];
        [ghostWindow orderOut:self];

    }
    else
    {
        // Decide whether this is the first click of a double click, or just a single click.
        // If there is another mouse down within 1/5 of a second, it's a double click.
        if ([NSApp nextEventMatchingMask:NSLeftMouseDownMask
                   untilDate:[NSDate dateWithTimeIntervalSinceNow:0.2]
                   inMode:NSModalPanelRunLoopMode
                   dequeue:NO] == nil)
        {
            unsigned int modifierFlags;

            // It looks like this will be a lone single click.
            modifierFlags = [theEvent modifierFlags];

            if (modifierFlags & NSShiftKeyMask)
            {
                [nonretainedDelegate toggleSelectionOfTile:self];
            }
            else
            {
                [nonretainedDelegate selectTile:self];
            }
        }

        else
        {
            // Double click.
            // Let the code below handle it.
        }

        // If we actually dragged, gobble up the mouse up event, otherwise pass it on.
        [aTileView mouseUp:theEvent];
    }
}

//----------------------------------------------------------------------

// XXX: inRect:ofView:?
- (NSEvent *) trackMouseInRect:(NSRect)rect
{
    NSPoint currentPoint;
    NSEvent *theEvent;

    do
    {
        theEvent = [tileWindow nextEventMatchingMask:(NSLeftMouseUpMask | NSLeftMouseDraggedMask)];
        currentPoint = [tileView convertPoint:[theEvent locationInWindow] fromView:nil];
    }
    while ([theEvent type] == NSLeftMouseDragged && NSPointInRect (currentPoint, rect) == YES);

    return theEvent;
}

//----------------------------------------------------------------------

- (void) tileView:(FITileView *)aTileView doubleClicked:(NSEvent *)theEvent
{
    unsigned int modifierFlags;
    BOOL hideOthers;
    
    modifierFlags = [theEvent modifierFlags];
    hideOthers = (modifierFlags & NSCommandKeyMask) ? YES : NO;
    [self launchApplicationAndHideOthers:hideOthers];
}

//----------------------------------------------------------------------

- (void) tileView:(FITileView *)aTileView acceptedFiles:(NSArray *)filenames
{
    if ([self isRunning] == NO)
        [self enterLaunching1State];

    [[FILauncher sharedLauncher] openFiles:filenames withApplication:[application path]];
}

//======================================================================
// Debugging
//======================================================================

- (NSMutableDictionary *) debugDictionary
{
    NSMutableDictionary *debugDictionary;

    debugDictionary = [super debugDictionary];
    if (application != nil)
        [debugDictionary setObject:application forKey:@"application"];
    [debugDictionary setObject:NSStringFromBool (hasLoadedFileIcon) forKey:@"hasLoadedFileIcon"];
    [debugDictionary setObject:NSStringFromBool (isSelected) forKey:@"isSelected"];
    [debugDictionary setObject:NSStringFromApplicationState (state) forKey:@"state"];

    return debugDictionary;
}

//======================================================================
// Handling Workspace notifications
//======================================================================

- (void) applicationWillLaunch:(NSNotification *)aNotification
{
    NSString *launchedApplication;

    launchedApplication = [[aNotification userInfo] objectForKey:@"NSApplicationName"];
    if ([[application path] isEqual:launchedApplication] == YES)
    {
        //NSLog (@"%@ will launch.", [application path]);
        [self enterLaunching1State];
    }
}

//----------------------------------------------------------------------

- (void) applicationDidLaunch:(NSNotification *)aNotification
{
    NSString *launchedApplication;

    launchedApplication = [[aNotification userInfo] objectForKey:@"NSApplicationName"];
    if ([[application path] isEqual:launchedApplication] == YES)
    {
        //NSLog (@"%@ did launch.", [application path]);
        [self enterLaunching2State];
    }
}

//----------------------------------------------------------------------

- (void) applicationDidTerminate:(NSNotification *)aNotification
{
    NSString *launchedApplication;

    launchedApplication = [[aNotification userInfo] objectForKey:@"NSApplicationName"];
    if ([[application path] isEqual:launchedApplication] == YES)
    {
        //NSLog (@"%@ did terminate.", [application path]);
        [self enterInactiveState];
    }
}

//======================================================================
// Controlling the file icon
//======================================================================

- (BOOL) hasLoadedFileIcon
{
    return hasLoadedFileIcon;
}

//----------------------------------------------------------------------

- (NSImage *) fileIcon
{
    [self updateFileIconIfNeeded];
    return fileIcon;
}

//----------------------------------------------------------------------

- (void) updateFileIcon
{
    SNRelease (fileIcon);
    {
        fileIcon = [[[FILauncher sharedLauncher] iconForFile:[application path]] retain];
        // Conflicts with launching...
        //[[[FILauncher sharedLauncher] workspaceLock] lock];
        //fileIcon = [[[NSWorkspace sharedWorkspace] iconForFile:[application path]] retain];
        //[[[FILauncher sharedLauncher] workspaceLock] unlock];
    }
    [tileView setImage:fileIcon];
    // We may need to generate a notification, so the inspector sees the change.
    [[NSNotificationCenter defaultCenter] postNotificationName:FITileIconDidChangeNotification object:self];

    hasLoadedFileIcon = YES;
}

//----------------------------------------------------------------------

- (void) updateFileIconIfNeeded
{
    if (hasLoadedFileIcon == NO)
        [self updateFileIcon];
}

//======================================================================
// Controlling the tile window
//======================================================================

- (FITileWindow *) tileWindow
{
    return tileWindow;
}

//----------------------------------------------------------------------

- (BOOL) isVisible
{
    return isVisible;
}

//----------------------------------------------------------------------

- (void) setIsVisible:(BOOL)flag
{
    if (flag == isVisible)
        return;

    isVisible = flag;
    if (isVisible == YES)
    {
        [self updateFileIconIfNeeded];
        [tileWindow orderFront:self];
    }
    else
    {
        [tileWindow orderOut:self];
    }
}

//----------------------------------------------------------------------

- (void) hide
{
    [self setIsVisible:NO];
}

//----------------------------------------------------------------------

- (void) unhide
{
    [self setIsVisible:YES];
}

//----------------------------------------------------------------------

- (void) repositionRelativeToPoint:(NSPoint)origin
{
    float tileWidth;
    DockLocation location;

    OBASSERT (application != nil);
    location = [application location];
    tileWidth = [[OFUserDefaults sharedUserDefaults] floatForKey:DK_TileSize];
    origin.x += location.x * tileWidth;
    origin.y += location.y * tileWidth;
    [tileWindow setFrameOrigin:origin];
}

//======================================================================
// NSWindow delegate
//======================================================================

- (NSUndoManager *) windowWillReturnUndoManager:(NSWindow *)window
{
    return [application undoManager];
}

//======================================================================
// Handling model notifications
//======================================================================

- (void) applicationDidChange:(NSNotification *)aNotification
{
    [tileView setNeedsDisplay:YES];

    if (animationImage == nil && [self shouldAnimate] == YES)
        [self startAnimationTimer];
    else if (animationImage != nil && [self shouldAnimate] == NO)
        [self stopAnimationTimer];
}

//----------------------------------------------------------------------

- (void) applicationDidChangeDockLocation:(NSNotification *)aNotification
{
    [nonretainedDelegate repositionTile:self];
}

//======================================================================
// Other
//======================================================================

- (void) checkForActiveApplications:(NSDictionary *)activeApplications
{
    FIAppInfo *anAppInfo;

    OBPRECONDITION (activeApplications != nil);

    if (hasCheckedIfActive == NO)
    {
        anAppInfo = [activeApplications objectForKey:[application shortName]];
        if (anAppInfo != nil)
        {
            [self enterActiveState:anAppInfo];
        }

        hasCheckedIfActive = YES;
    }
}

//----------------------------------------------------------------------

// Autolaunching -- only activate app if fiend is active
// Otherwise -- activate launched app when is has successfully launched
// - forget weird "last launch path" stuff?

- (void) launchApplicationAndHideOthers:(BOOL)hideOthers
{
    if ([self isRunning] == NO)
    {
        tileLaunchedApplication = YES;
        [self enterLaunching1State];
        [tileView setNeedsDisplay:YES];

        // Make sure.
        [self stopAutoLaunchTimer];

        [self enterLaunching1State]; // XXX: New.
        [[FILauncher sharedLauncher] launchApplication:[application path] autolaunch:NO];
#if 0
        didLaunch = [[NSWorkspace sharedWorkspace] launchApplication:path showIcon:YES autolaunch:NO];
        if (didLaunch == NO)
        {
            [self enterInactiveState];
        }
        else
        {
            // Activate the application.
            //NSLog (@"We want to activate, appInfo: %@", appInfo);
            //[NSApp hide:self];
        }
#endif
    }
    else
    {
        // Activate the application.
        if (appInfo != nil)
        {
            [appInfo activateApplicationAndHideOthers:hideOthers];
        }
        else
        {
            NSBeep ();
        }
    }
}

//----------------------------------------------------------------------

- (void) performAutoLaunch
{
    if ([self isRunning] == NO && [self shouldAutoLaunch] == YES)
    {
        float delaySeconds;
#if 0
        NSLog (@"need to auto launch: %@ with delay: %@",
               [application path], NSStringFromTimeInterval ([application autoLaunchDelay]));
#endif
        delaySeconds = secondsInTimeInterval ([application autoLaunchDelay]);
        if (delaySeconds == 0)
        {
            // XXX: Rename autolaunch -> autoLaunch.
            [self enterLaunching1State]; // XXX: New.
            [[FILauncher sharedLauncher] launchApplication:[application path] autolaunch:YES];
        }
        else
        {
            [self startAutoLaunchTimer];
        }
    }
}

@end
