#import "Fiend.h"

RCS_ID ("$Id$");

#import "FIDock.h"

#import "FIApplication.h"
#import "FILevel.h"

DEFINE_NSSTRING (FIDockDidAddLevelNotification);
DEFINE_NSSTRING (FIDockDidRemoveLevelNotification);
DEFINE_NSSTRING (FIDockDidMoveApplicationBetweenLevelsNotification);

@implementation FIDock

- init
{
    FILevel *firstLevel;

    if ([super init] == nil)
        return nil;

    persistentLevel = [[FILevel allocWithZone:[self zone]] init];
    [persistentLevel setIsPersistent:YES];
    [persistentLevel setDock:self];

    levels = [[NSMutableArray allocWithZone:[self zone]] initWithCapacity:4];
    firstLevel = [[FILevel allocWithZone:[self zone]] init];
    [firstLevel setDock:self];
    [levels addObject:firstLevel];
    [firstLevel release];

    return self;
}

//----------------------------------------------------------------------

- (void) dealloc
{
    SNRelease (levels);
    SNRelease (persistentLevel);

    [super dealloc];
}

//======================================================================
// Loading and saving
//======================================================================

//" This should only be called on a newly created (empty) dock.
- (void) loadDockFromPropertyList:(NSDictionary *)propertyList
{
    NSDictionary *savedLevelPropertyList;
    NSArray *loadedLevelsPropertyList;

    savedLevelPropertyList = [propertyList objectForKey:@"PersistentDockLevel"];
    if (savedLevelPropertyList != nil)
    {
        [persistentLevel loadLevelFromPropertyList:savedLevelPropertyList];
        //SNLog (persistentLevel);
    }

    loadedLevelsPropertyList = [propertyList objectForKey:@"DockLevels"];
    if (loadedLevelsPropertyList != nil)
    {
        NSEnumerator *levelEnumerator;
        BOOL isFirstLevel;
        FILevel *newLevel;

        isFirstLevel = YES;
        levelEnumerator = [loadedLevelsPropertyList objectEnumerator];
        while (savedLevelPropertyList = [levelEnumerator nextObject])
        {
            if (isFirstLevel == YES)
            {
                OBPRECONDITION ([levels count] > 0);
                newLevel = [levels objectAtIndex:0];
            }
            else
            {
                newLevel = [[[FILevel allocWithZone:[self zone]] init] autorelease];
                [newLevel setDock:self];
                [levels addObject:newLevel];
            }
            
            [newLevel loadLevelFromPropertyList:savedLevelPropertyList];
            //SNLog (newLevel);
            //[newLevel loadLevelFromPropertyList:[savedLevelPropertyList objectForKey:@"DockedApplications"]];
            //[newLevel setLevelName:[savedLevelPropertyList objectForKey:@"DockLevelName"]];
            isFirstLevel = NO;
        }
    }
}

//----------------------------------------------------------------------

- (NSDictionary *) propertyListForDock
{
    NSMutableDictionary *savedDockPropertyList;
    NSMutableArray *savedLevelsPropertyList;
    NSEnumerator *levelEnumerator;
    FILevel *level;

    savedDockPropertyList = [NSMutableDictionary dictionary];
    [savedDockPropertyList setObject:[persistentLevel propertyListForLevel] forKey:@"PersistentDockLevel"];

    savedLevelsPropertyList = [NSMutableArray array];
    levelEnumerator = [levels objectEnumerator];
    while (level = [levelEnumerator nextObject])
    {
        [savedLevelsPropertyList addObject:[level propertyListForLevel]];
    }

    [savedDockPropertyList setObject:savedLevelsPropertyList forKey:@"DockLevels"];

    return savedDockPropertyList;
}     

//======================================================================
// Controlling the document
//======================================================================

- (void) setDocument:(NSDocument *)newDocument
{
    [super setDocument:newDocument];
    [persistentLevel setDocument:newDocument];
    [levels makeObjectsPerformSelector:@selector (setDocument:) withObject:newDocument];
}

//======================================================================
// Controlling levels
//======================================================================

- (FILevel *) persistentLevel
{
    return persistentLevel;
}

//----------------------------------------------------------------------

- (NSArray *) levels
{
    return levels;
}

//----------------------------------------------------------------------

- (int) indexOfLevel:(FILevel *)level
{
    return [levels indexOfObject:level];
}

//----------------------------------------------------------------------

- (NSString *) displayNameOfLevel:(FILevel *)level
{
    NSString *name;

    name = [level levelName];
    if (name == nil)
    {
        int count, index;

        count = [levels count];
        index = [levels indexOfObject:level];

        if (index == NSNotFound)
            name = @"";
        else
            name = [NSString stringWithFormat:@"%d-%d", index + 1, count];
    }

    return name;
}

//----------------------------------------------------------------------

- (void) addLevel:(FILevel *)level
{
    [self addLevel:level atIndex:[levels count]];
}

//----------------------------------------------------------------------

- (void) addLevel:(FILevel *)level atIndex:(int)index
{
    NSDictionary *userInfo;
    int count;

    count = [levels count];
    OBASSERT (index >= 0 && index <= count);
    // if index == count?
    [[[self undoManager] prepareWithInvocationTarget:self] removeLevel:level];
    [levels insertObject:level atIndex:index];
    [level setDock:self];
    [level setDocument:[self document]];
    userInfo = [NSDictionary dictionaryWithObject:level forKey:@"level"];
    [[NSNotificationCenter defaultCenter] postNotificationName:FIDockDidAddLevelNotification object:self userInfo:userInfo];
}

//----------------------------------------------------------------------

- (void) removeLevel:(FILevel *)level
{
    NSDictionary *userInfo;
    OBPRECONDITION ([levels containsObject:level]);
    [[[self undoManager] prepareWithInvocationTarget:self] addLevel:level atIndex:[levels indexOfObject:level]];
    [levels removeObject:level];
    userInfo = [NSDictionary dictionaryWithObject:level forKey:@"level"];
    [[NSNotificationCenter defaultCenter] postNotificationName:FIDockDidRemoveLevelNotification object:self userInfo:userInfo];
}

//----------------------------------------------------------------------

// removing from one, and then adding to another, causes the tile to flash out, then in.

- (void) moveApplication:(FIApplication *)application toLevel:(FILevel *)newLevel
{
    FILevel *oldLevel;
    NSDictionary *userInfo;

    OBPRECONDITION (newLevel != nil);

    oldLevel = [application level];
    OBASSERT (oldLevel != nil);

    [[[self undoManager] prepareWithInvocationTarget:self] moveApplication:application toLevel:oldLevel];

    [application retain];
    [oldLevel _removeApplication:application];
    [newLevel _addApplication:application];

    [application setIsPersistent:[newLevel isPersistent]];
    [application release];

    userInfo = [NSDictionary dictionaryWithObjects:[NSArray arrayWithObjects:application, oldLevel, nil]
                             forKeys:[NSArray arrayWithObjects:@"application", @"oldLevel", nil]];

    [[NSNotificationCenter defaultCenter] postNotificationName:FIDockDidMoveApplicationBetweenLevelsNotification
                                          object:self
                                          userInfo:userInfo];
}

//======================================================================
// Debugging
//======================================================================

- (NSMutableDictionary *) debugDictionary
{
    NSMutableDictionary *debugDictionary;

    debugDictionary = [super debugDictionary];

    if (persistentLevel != nil)
        [debugDictionary setObject:persistentLevel forKey:@"persistentLevel"];
    if (levels != nil)
        [debugDictionary setObject:levels forKey:@"levels"];

    return debugDictionary;
}

//======================================================================
// Other
//======================================================================

- (BOOL) containsApplicationWithPath:(NSString *)applicationPath
{
    BOOL isDocked;

    isDocked = [persistentLevel containsApplicationWithPath:applicationPath];
    if (isDocked == NO)
    {
        NSEnumerator *levelEnumerator;
        FILevel *level;

        levelEnumerator = [levels objectEnumerator];
        while (level = [levelEnumerator nextObject])
        {
            isDocked = [level containsApplicationWithPath:applicationPath];
            if (isDocked == YES)
                break;
        }
    }

    return isDocked;
}

//----------------------------------------------------------------------

- (void) removeApplicationsFromArray:(NSArray *)applications
{
    NSEnumerator *applicationEnumerator;
    FIApplication *application;

    applicationEnumerator = [applications objectEnumerator];
    while (application = [applicationEnumerator nextObject])
    {
        [[application level] removeApplication:application];
    }
}

//----------------------------------------------------------------------

// If start is nil, begin with persistent level
// Otherwise, start at level of start app.

- (FIApplication *) findNextApplication:(FIApplication *)startingApplication
                             withString:(NSString *)searchString
                             ignoreCase:(BOOL)ignoreCase
{
    FIApplication *nextApplication = nil;
    int count, l;

    if (startingApplication == nil)
    {
        nextApplication = [persistentLevel findNextApplication:nil withString:searchString ignoreCase:ignoreCase];
        if (nextApplication == nil)
        {
            count = [levels count];
            for (l = 0; l < count; l++)
            {
                nextApplication = [[levels objectAtIndex:l] findNextApplication:nil
                                                            withString:searchString
                                                            ignoreCase:ignoreCase];
                if (nextApplication != nil)
                    break;
            }
        }
    }
    else
    {
        FILevel *currentLevel;

        currentLevel = [startingApplication level];
        nextApplication = [currentLevel findNextApplication:startingApplication
                                        withString:searchString
                                        ignoreCase:ignoreCase];
        if (nextApplication == nil)
        {
            int startingLevelIndex;

            if (currentLevel == persistentLevel)
                startingLevelIndex = 0;
            else
                startingLevelIndex = [levels indexOfObject:currentLevel] + 1;

            count = [levels count];
            for (l = startingLevelIndex; l < count; l++)
            {
                nextApplication = [[levels objectAtIndex:l] findNextApplication:nil
                                                            withString:searchString
                                                            ignoreCase:ignoreCase];
                if (nextApplication != nil)
                    break;
            }
        }
    }

    return nextApplication;
}

//----------------------------------------------------------------------

- (FIApplication *) findPreviousApplication:(FIApplication *)startingApplication
                                 withString:(NSString *)searchString
                                 ignoreCase:(BOOL)ignoreCase
{
    FIApplication *previousApplication = nil;
    int count, l;

    if (startingApplication == nil)
    {
        count = [levels count];
        for (l = count - 1; l >= 0; l--)
        {
            previousApplication = [[levels objectAtIndex:l] findPreviousApplication:nil
                                                            withString:searchString
                                                            ignoreCase:ignoreCase];
            if (previousApplication != nil)
                break;
        }

        if (previousApplication == nil)
        {
            previousApplication = [persistentLevel findPreviousApplication:nil withString:searchString ignoreCase:ignoreCase];
        }
    }
    else
    {
        FILevel *currentLevel;

        currentLevel = [startingApplication level];
        previousApplication = [currentLevel findPreviousApplication:startingApplication
                                            withString:searchString
                                            ignoreCase:ignoreCase];
        if (previousApplication == nil)
        {
            int startingLevelIndex;

            if (currentLevel != persistentLevel)
            {
                startingLevelIndex = [levels indexOfObject:currentLevel] - 1;

                count = [levels count];
                for (l = startingLevelIndex; l >= 0; l--)
                {
                    previousApplication = [[levels objectAtIndex:l] findPreviousApplication:nil
                                                                    withString:searchString
                                                                    ignoreCase:ignoreCase];
                    if (previousApplication != nil)
                        break;
                }

                if (previousApplication == nil)
                {
                    previousApplication = [persistentLevel findPreviousApplication:nil
                                                           withString:searchString
                                                           ignoreCase:ignoreCase];
                }
            }
        }
    }

    return previousApplication;
}

//----------------------------------------------------------------------

- (FILevel *) findNextLevel:(FILevel *)startingLevel withString:(NSString *)searchString ignoreCase:(BOOL)ignoreCase
{
    int startIndex, count, l;
    FILevel *level, *nextLevel = nil;
    NSRange searchRange;
    unsigned int options;

    OBPRECONDITION (startingLevel != nil);

    options = (ignoreCase == YES) ? NSCaseInsensitiveSearch : 0;

    startIndex = [levels indexOfObject:startingLevel];
    OBASSERT (startIndex != NSNotFound);
    startIndex++;

    count = [levels count];
    for (l = startIndex; l < count; l++)
    {
        level = [levels objectAtIndex:l];
        searchRange = [[level displayName] rangeOfString:searchString options:options];
        if (searchRange.length != 0)
        {
            nextLevel = level;
            break;
        }
    }

    return nextLevel;
}

//----------------------------------------------------------------------
    
- (FILevel *) findPreviousLevel:(FILevel *)startingLevel withString:(NSString *)searchString ignoreCase:(BOOL)ignoreCase
{
    int startIndex, count, l;
    FILevel *level, *previousLevel = nil;
    NSRange searchRange;
    unsigned int options;

    OBPRECONDITION (startingLevel != nil);

    options = (ignoreCase == YES) ? NSCaseInsensitiveSearch : 0;

    startIndex = [levels indexOfObject:startingLevel];
    OBASSERT (startIndex != NSNotFound);
    startIndex--;

    count = [levels count];
    for (l = startIndex; l >= 0; l--)
    {
        level = [levels objectAtIndex:l];
        searchRange = [[level displayName] rangeOfString:searchString options:options];
        if (searchRange.length != 0)
        {
            previousLevel = level;
            break;
        }
    }

    return previousLevel;
}

@end
