#import "Fiend.h"

RCS_ID ("$Id$");

#import "FILevel.h"

#import "FIApplication.h"
#import "FIDock.h"

DEFINE_NSSTRING (FILevelNameDidChangeNotification);
DEFINE_NSSTRING (FILevelDidAddApplicationNotification);
DEFINE_NSSTRING (FILevelDidRemoveApplicationNotification);

@implementation FILevel

- init
{
    if ([super init] == nil)
        return nil;

    nonretainedDock = nil;
    levelName = nil;
    applications = [[NSMutableArray allocWithZone:[self zone]] initWithCapacity:8];
    isPersistent = NO;

    return self;
}

//----------------------------------------------------------------------

- (void) dealloc
{
    SNRelease (applications);
    SNRelease (levelName);

    [super dealloc];
}

//======================================================================
// Loading and saving
//======================================================================

//" This should only be called on a newly created (empty) level.
- (void) loadLevelFromPropertyList:(NSDictionary *)savedLevelPropertyList
{
    NSArray *savedApplications;
    NSEnumerator *savedApplicationEnumerator;
    NSDictionary *savedApplication;
    FIApplication *newApplication;
    
    [self setLevelName:[savedLevelPropertyList objectForKey:@"DockLevelName"]];
    savedApplications = [savedLevelPropertyList objectForKey:@"DockedApplications"];

    savedApplicationEnumerator = [savedApplications objectEnumerator];
    while (savedApplication = [savedApplicationEnumerator nextObject])
    {
        newApplication = [FIApplication applicationFromPropertyList:savedApplication];
        if (newApplication != nil)
        {
            [newApplication setIsPersistent:isPersistent];
            [self addApplication:newApplication];
        }
    }
}

//----------------------------------------------------------------------

- (NSDictionary *) propertyListForLevel
{
    NSMutableDictionary *propertyList;
    NSMutableArray *applicationList;
    NSEnumerator *applicationEnumerator;
    FIApplication *application;

    propertyList = [NSMutableDictionary dictionary];
    applicationList = [NSMutableArray array];

    applicationEnumerator = [applications objectEnumerator];
    while (application = [applicationEnumerator nextObject])
    {
        [applicationList addObject:[application propertyListForApplication]];
    }

    if (levelName != nil)
    {
        [propertyList setObject:levelName forKey:@"DockLevelName"];
    }
    [propertyList setObject:applicationList forKey:@"DockedApplications"];

    return propertyList;
}


//======================================================================
// Controlling the document
//======================================================================

- (void) setDocument:(NSDocument *)newDocument
{
    [super setDocument:newDocument];
    [applications makeObjectsPerformSelector:@selector (setDocument:) withObject:newDocument];
}

//======================================================================
// Controlling general attributes
//======================================================================

- (FIDock *) dock
{
    return nonretainedDock;
}

//----------------------------------------------------------------------

- (void) setDock:(FIDock *)newDock
{
    if (newDock == nonretainedDock)
        return;

    nonretainedDock = newDock;
}

//----------------------------------------------------------------------

- (NSString *) levelName
{
    return levelName;
}

//----------------------------------------------------------------------

- (void) setLevelName:(NSString *)newName
{
    // XXX: Does this handle nils?
    if (newName == levelName || [newName isEqual:levelName] == YES)
        return;

    [[[self undoManager] prepareWithInvocationTarget:self] setLevelName:levelName];
    SNRelease (levelName);
    levelName = [newName copyWithZone:[self zone]];
    [[NSNotificationCenter defaultCenter] postNotificationName:FILevelNameDidChangeNotification object:self];
}

//----------------------------------------------------------------------

- (NSString *) displayName
{
    NSString *displayName;
    //return [nonretainedDock displayNameOfLevel:self];
    displayName = [nonretainedDock displayNameOfLevel:self];
    return displayName;
}

//----------------------------------------------------------------------

- (NSString *) displayName:sender
{
    return [self displayName];
}

//----------------------------------------------------------------------

- (BOOL) isPersistent
{
    return isPersistent;
}

//----------------------------------------------------------------------

- (void) setIsPersistent:(BOOL)flag
{
    if (flag == isPersistent)
        return;

    [[[self undoManager] prepareWithInvocationTarget:self] setIsPersistent:isPersistent];
    isPersistent = flag;
}

//======================================================================
// Controlling applications
//======================================================================

- (NSArray *) applications
{
    return applications;
}

//----------------------------------------------------------------------

- (BOOL) containsApplicationWithPath:(NSString *)applicationPath
{
    NSEnumerator *applicationEnumerator;
    FIApplication *application;
    BOOL containsApplication;

    containsApplication = NO;

    applicationEnumerator = [applications objectEnumerator];
    while (application = [applicationEnumerator nextObject])
    {
        if ([[application path] isEqual:applicationPath] == YES)
        {
            containsApplication = YES;
            break;
        }
    }

    return containsApplication;
}

//----------------------------------------------------------------------

// These two are used by FIDock when moving applications between levels.

- (void) _addApplication:(FIApplication *)newApplication
{
    [applications addObject:newApplication];
    [newApplication setLevel:self];
    [newApplication setDocument:[self document]];
}

//----------------------------------------------------------------------

- (void) _removeApplication:(FIApplication *)anApplication
{
    [anApplication setLevel:nil];
    [applications removeObject:anApplication];
}

//----------------------------------------------------------------------

- (void) addApplication:(FIApplication *)newApplication
{
    [[[self undoManager] prepareWithInvocationTarget:self] removeApplication:newApplication];

    [self _addApplication:newApplication];
    [[NSNotificationCenter defaultCenter] postNotificationName:FILevelDidAddApplicationNotification
                                          object:self
                                          userInfo:[NSDictionary dictionaryWithObject:newApplication forKey:@"application"]];
}

//----------------------------------------------------------------------

- (void) removeApplication:(FIApplication *)anApplication
{
    [[[self undoManager] prepareWithInvocationTarget:self] addApplication:anApplication];

    [anApplication retain];
    [self _removeApplication:anApplication];
    [[NSNotificationCenter defaultCenter] postNotificationName:FILevelDidRemoveApplicationNotification
                                          object:self
                                          userInfo:[NSDictionary dictionaryWithObject:anApplication forKey:@"application"]];
    [anApplication release];
}

//======================================================================
// Controlling locations
//======================================================================

- (BOOL) isDockLocationEmpty:(DockLocation)location
{
    NSEnumerator *applicationEnumerator;
    FIApplication *application;
    DockLocation currentLocation;
    BOOL isEmpty;

    isEmpty = YES;

    applicationEnumerator = [applications objectEnumerator];
    while (application = [applicationEnumerator nextObject])
    {
        currentLocation = [application location];
        if (currentLocation.x == location.x && currentLocation.y == location.y)
        {
            isEmpty = NO;
            break;
        }
    }

    return isEmpty;
}

//----------------------------------------------------------------------

- (void) rotateApplicationsClockwise:(BOOL)clockwise
{
    NSEnumerator *applicationEnumerator;
    FIApplication *application;

    applicationEnumerator = [applications objectEnumerator];
    while (application = [applicationEnumerator nextObject])
    {
        [application rotateLocationClockwise:clockwise];
    }
}

//======================================================================
// Debugging
//======================================================================

- (NSMutableDictionary *) debugDictionary
{
    NSMutableDictionary *debugDictionary;

    debugDictionary = [super debugDictionary];
    if (nonretainedDock != nil)
        [debugDictionary setObject:[nonretainedDock shortDescription] forKey:@"nonretainedDock"];
    if (levelName != nil)
        [debugDictionary setObject:levelName forKey:@"levelName"];
    if (applications != nil)
        [debugDictionary setObject:applications forKey:@"applications"];
    [debugDictionary setObject:NSStringFromBool (isPersistent) forKey:@"isPersistent"];

    return debugDictionary;
}

//======================================================================
// Other
//======================================================================

- (void) clearLevel
{
    NSArray *oldApplications;
    NSEnumerator *applicationEnumerator;
    FIApplication *application;

    oldApplications = [NSArray arrayWithArray:applications];
    applicationEnumerator = [oldApplications objectEnumerator];
    while (application = [applicationEnumerator nextObject])
    {
        [self removeApplication:application];
    }
}

//----------------------------------------------------------------------

- (FIApplication *) findNextApplication:(FIApplication *)startingApplication
                             withString:(NSString *)searchString
                             ignoreCase:(BOOL)ignoreCase
{
    int start, count, l;
    FIApplication *application, *targetApplication = nil;
    NSRange searchRange;
    unsigned int options;

    count = [applications count];

    if (startingApplication == nil)
    {
        start = 0;
    }
    else
    {
        start = [applications indexOfObject:startingApplication];
        OBASSERT (start != NSNotFound);
        start++; // Skip over current application
    }

    options = (ignoreCase == YES) ? NSCaseInsensitiveSearch : 0;
    for (l = start; l < count; l++)
    {
        application = [applications objectAtIndex:l];
        searchRange = [[application path] rangeOfString:searchString options:options];
        if (searchRange.length != 0)
        {
            targetApplication = application;
            break;
        }
    }

    return targetApplication;
}

//----------------------------------------------------------------------

- (FIApplication *) findPreviousApplication:(FIApplication *)startingApplication
                                 withString:(NSString *)searchString
                                 ignoreCase:(BOOL)ignoreCase
{
    int start, count, l;
    FIApplication *application, *targetApplication = nil;
    NSRange searchRange;
    unsigned int options;

    count = [applications count];

    if (startingApplication == nil)
    {
        start = count - 1;
    }
    else
    {
        start = [applications indexOfObject:startingApplication];
        OBASSERT (start != NSNotFound);
        start--; // Skip over current application
    }

    options = (ignoreCase == YES) ? NSCaseInsensitiveSearch : 0;
    for (l = start; l >= 0; l--)
    {
        application = [applications objectAtIndex:l];
        searchRange = [[application path] rangeOfString:searchString options:options];
        if (searchRange.length != 0)
        {
            targetApplication = application;
            break;
        }
    }

    return targetApplication;
}

@end
